(* Delphi Unit
   several mathematical functions
   ==============================

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jun. 1989
     last changed: Nov. 2001 
   *)

unit WinMath;

interface

uses System.SysUtils, System.Math;

const
  Pi180 = Pi/180.0;   (* Grad -> Bogenma *)
  Pi200 = Pi/200.0;   (* Neugrad -> Bogenma *)

type
  EinhStr = string[5];
  TAngleFormat = (fmDezGrad,fmDezMin,fmSekunden);

  TMean = class (TObject)
    Count : cardinal;
    Sum   : double;
    constructor Create;
    procedure Reset;
    procedure Add(Value : integer); overload;
    procedure Add(Value : double);  overload;
    function Mean: double;
    end;

var
  ln2,ln10,sqrt2 : extended;

function Pwr2 (exp : integer) : int64;

function AInt (x : double) : double;
function IInt (x : double) : integer;

function RMod (x,y : double) : double;

function Sign (x,y : double) : double;

function Lg (x : double) : double;

function Ld (x : double) : double;

function Pwr (x,y : double) : double;

function Kub (x : double) : double;

function Tan (x : double) : double;

function ArcTan1 (x,y : double) : double;

function ArcTan2 (x,y : double) : double;

function ArcTanB (x,y : double) : double;

function ISign (x,y : integer) : integer;

function RMin (x,y : double) : double;

function RMax (x,y : double) : double;

(* Typumwandlung prfen *)
function CheckValue (s     : string) : boolean;

(* an Dezimalstelle runden *)
function DecimalRound (Value : double; Decimal : integer) : double;

(* Normalisieren eines Wertes in den Bereich 0.1 .. 999.9 *)
procedure Normal (var Value,Factor : double;
                  var PrefixIndex  : integer);

(* Zahlenwert mit nachfolgendem Einheitenvorsatz erzeugen *)
function PrefixVal (Value : double;
                    n,nd  : integer;
                    Einh  : EinhStr) : string;

(* Zahlenwert aus String mit nachfolgendem Einheitenvorsatz erzeugen *)
function ValPrefix (ValStr    : string;
                    var Value : double) : boolean;

(* Zahlen auf "Sci-Format" umsetzen *)
function FloatToStrS (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;

(* Zahlen auf "Eng-Format" umsetzen *)
function FloatToStrE (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;

(* Zahlen auf "Eng-Format" mit Vorstzen (m,k,..)  umsetzen *)
function FloatToPrefixStr (Value     : double;
                           DecSep    : char;
                           n         : integer;
                           Separator : string) : string;

(* Zahlen auf "ffXixed-Format" umsetzen *)
function FloatToFixStr (Value     : double;
                        DecSep  : char;
                        n         : integer) : string;

(* In Winkelgrade umrechnen *)
function DegToStr(v : double; Format : TAngleFormat; DecSep : char; w,d : integer) : string;
function StrToDeg (s : string; var v : double) : boolean;

implementation

const
  PfAnz = 10;
  Pfx   : array [1..PfAnz] of char = ('f','p','n','','m',' ','k','M','G','T');
  PfFak : array [0..PfAnz] of double = (1.0,1E15,1E12,1E9,1E6,1E3,
                                        1.0,1E-3,1E-6,1E-9,1E-12);
  FFak : array [-2..8] of double = (0.01,0.1,1.0,10.0,100.0,1000.0,1E4,1E5,1E6,1E7,1E8);
  NFak : array [1..2] of integer = (10,100);

{ ------------------------------------------------------------------- }
// Compute mean value
constructor TMean.Create;
begin
  inherited Create;
  Reset;
  end;

procedure TMean.Reset;
begin
  Count:=0; Sum:=0;
  end;

procedure TMean.Add(Value : integer);
begin
  inc(Count); Sum:=Sum+Value;
  end;

procedure TMean.Add(Value : double);
begin
  inc(Count); Sum:=Sum+Value;
  end;

function TMean.Mean: double;
begin
  if Count=0 then Result:=0
  else Result:=Sum/Count;
  end;

{ ------------------------------------------------------------------- }
(* Compute power of 2 as integer *)
function Pwr2 (exp : integer) : int64;
begin
  if exp>63 then exp:=63;
  if exp<0 then exp:=0;
  Result:=1;
  Result:=Result shl exp;
//  for i:=1 to exp do Result:=2*Result;
  end;


(* ----------FUNCTION--AINT---------------------------------------------
   Bestimmung des nchstkleineren ganzzahligen Anteils einer Zahl
*)
function AInt (x : double) : double;
var
  xx  : double;
begin
  xx:=int(x);
  if (x<0.0) and (x<>xx) then xx:=xx-1.0;
  Result:=xx;
  end;

function IInt (x : double) : integer;
begin
  Result:=round(AInt(x));
  end;

(* ----------FUNCTION--RMOD---------------------------------------------
   Divisionsrest bei Fliekommazahlen
*)
function RMod (x,y : double) : double;
begin
  RMod:=x-int(1.000001*x/y)*y;  // Faktor erforderlich wegen Rundungsfehler
  end;

(* ----------FUNCTION--ISIGN--------------------------------------------
  (Vorzeichen von arg2) * arg1
*)
function ISign (x,y : integer) : integer;
begin
  if y<0 then isign:=-x else isign:=x; 
  end;

(* ----------FUNCTION--SIGN---------------------------------------------
  (Vorzeichen von arg2) * arg1
*)
function Sign (x,y : double) : double;
begin
  if y<0.0 then sign:=-x else if y>0.0 then sign:=x else sign:=0.0;
  end;

(* ----------FUNCTION--LG-----------------------------------------------
  Dekadischer Logarithmus
*)
function Lg (x : double) : double;
begin
  lg:=ln(x)/ln10;
  end;

(* ----------FUNCTION--Ld-----------------------------------------------
  Dualer Logarithmus
*)
function Ld (x : double) : double;
begin
  Ld:=ln(x)/ln2;
  end;

(* ----------FUNCTION--PWR----------------------------------------------
  allgemeine Exponentialfunktion x ^ y
*)
function Pwr (x,y : double) : double;
begin
  pwr:=exp (y*ln(x));
  end;

(* ----------FUNCTION--KUB----------------------------------------------
   Kubik (x hoch 3) *)
function Kub (x : double) : double;
begin
  Kub:=sqr(x)*x;
  end;

(* ----------FUNCTION--TAN----------------------------------------------
  Tangens
  *)
function Tan (x : double) : double;
var
  c : double;
begin
  c:=cos(x);
  if c<>0.0 then Tan:=sin(x)/c
  else Tan:=Maxdouble;
  end;

(* ----------FUNCTION--ArcTan1------------------------------------------
   ArcTan mit Quadrantenbestimmung, Winkel von 0 .. 360 Grad
   linksweisend gegen Ost (math. pos. Sinn) *)
function ArcTan1 (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=90.0-ArcTan (x/y)/Pi180;
    if y<0.0 then z:=z+180.0;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=ArcTan (y/x)/Pi180;
    if x<0.0 then z:=z+180.0;
    end;
  if z<0.0 then z:=z+360.0;
  ArcTan1:=z;
  end;

(* ----------FUNCTION--ArcTan2------------------------------------------
   ArcTan mit Quadrantenbestimmung, Winkel von 0 .. 360 Grad
   rechtsweisend gegen Nord
*)
function ArcTan2 (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=ArcTan (x/y)/Pi180;
    if y<0.0 then z:=z+180.0
    else if x<0.0 then z:=z+360.0;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=90.0-ArcTan (y/x)/Pi180;
    if x<0.0 then z:=z+180.0;
    end;
  ArcTan2:=z;
  end;

(* ----------FUNCTION--ArcTan------------------------------------------
   ArcTan mit Quadrantenbestimmung, Winkel in Bogenma
   +/- Pi im math. pos. Sinn *)
function ArcTanB (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=0.5*Pi-ArcTan (x/y);
    if y<0.0 then z:=z+Pi
    else if x<0.0 then z:=z+2*Pi;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=ArcTan (y/x);
    if x<0.0 then z:=z+Pi;
    end;
  if z<=-Pi then z:=z+2*Pi;
  if z>Pi then z:=z-2*Pi;
  ArcTanB:=z;
  end;

(* ----------FUNCTION--RMIN---------------------------------------------
   Mininum zweier Realzahlen
*)
function rmin (x,y : double) : double;
begin
  if x<y then rmin:=x else rmin:=y;
  end;

(* ----------FUNCTION--RMAX---------------------------------------------
   Maximum zweier Realzahlen
*)
function rmax (x,y : double) : double;
begin
  if x<y then rmax:=y else rmax:=x;
  end;

(* ----------PROCEDURE-CheckValue---------------------------------------- *)
function CheckValue (s : string) : boolean;
begin
  Result:=true;
  if length(s)>0 then begin
    try
      StrToFloat(s);
    except
      on EConvertError do Result:=false;
      end;
    end;
  end;

(* ----------FUNCTION DECIMALROUND------------------------------------------- *)
(* an Dezimalstelle runden *)
function DecimalRound (Value : double; Decimal : integer) : double;
var
  v : integer;
begin
  if Decimal<-2 then Decimal:=-2;
  if Decimal>8 then Decimal:=8;
  v:=round(abs(Value)*FFak[Decimal]);
  Result:=v/FFak[Decimal];
  if Value<0 then Result:=-Result;
  end;

(* ----------PROCEDURE-NORMAL------------------------------------------- *)
(* Normalisieren eines Wertes in den Bereich 0.1 .. 999.9 *)
procedure Normal (var Value,Factor : double;
                  var PrefixIndex  : integer);
var
  AVal : double;
  k    : integer;
begin
  AVal:=abs(Value);
  if (AVal<1E-15) or (AVal>1E15) then k:=6
  else begin
    k:=1;
    if (AVal>=0.999E-13) then inc(k);
    if (AVal>=0.999E-10) then inc(k);
    if (AVal>=0.999E-7) then inc(k);
    if (AVal>=0.999E-4) then inc(k);
    if (AVal>=0.999E-1) then inc(k);
    if (AVal>=0.999E3) then inc(k);
    if (AVal>=0.999E6) then inc(k);
    if (AVal>=0.999E9) then inc(k);
    if (AVal>=0.999E12) then inc(k);
    end;
  Factor:=PfFak[k];
  Value:=Value*Factor;
  PrefixIndex:=k;
  end;

(* ----------FUNCTION--PREFIXVAL---------------------------------------- *)
(* Zahlenwert mit nachfolgendem Einheitenvorsatz erzeugen *)
function PrefixVal (Value   : double;
                    n,nd    : integer;
                    Einh    : EinhStr) : string;
(* n    = Anzahl der Stellen
        = 0: automatische Stellenanpassung
   nd   = Anzahl der Stellen hinter dem Komma
          <0 : automatisch max. -(nd+1) Stellen hinter dem Komma
   Einh = nachfolgendes Einheitenzeichen *)
var
  s    : string;
  pf   : char;
  f    : double;
  i    : integer;
begin
  Normal (Value,f,i); pf:=Pfx[i];
  if nd<0 then begin
    nd:=abs(nd)-2;
    if abs(Value)<100.0 then inc(nd);
    if abs(Value)<10.0 then inc(nd);
    if abs(Value)<1.0 then inc(nd);
    if abs(Value)<0.0999 then nd:=0;
    if nd<0 then nd:=0;
    end;
  if n=0 then begin
    n:=3+nd; if nd>0 then inc(n);
    if abs(Value)<100.0 then dec(n);
    if abs(Value)<10.0 then dec(n);
    end;
  if Value<0.0 then inc(n);
  s:=FloatToStrF(Value,ffFixed,n,nd);
  if Einh='' then begin
    if pf<>' ' then PrefixVal:=s+pf
    else PrefixVal:=s;
    end
  else if pf<>' ' then PrefixVal:=s+' '+pf+Einh
    else PrefixVal:=s+' '+Einh+' ';
  end;

(* ----------FUNCTION--VALPREFIX---------------------------------------- *)
(* Zahlenwert aus String mit nachfolgendem Einheitenvorsatz erzeugen *)
function ValPrefix (ValStr  : string;
                    var Value : double) : boolean;
(* PrefixVal = true : ok
             = false : Fehler bei der Konvertierung *)
var
  k,n  : integer;
  c    : char;
begin
  Result:=false;
  ValStr:=Trim(ValStr);
  (* Prfe letztes Zeichen *)
  n:=length(ValStr);
  if n>0 then begin
    c:=ValStr[n]; Value:=0.0; k:=0;
    if (c<'0') or (c>'9') then begin
      k:=1;
      while (k<=PfAnz) and (Pfx[k]<>c) do inc(k);
      end;
    if k<=PfAnz then begin
      if k>0 then begin
        dec(n);
        while (n>0) and (ValStr[n]=' ') do dec(n);
        end;
      if TryStrToFloat(copy(ValStr,1,n),Value) then begin
        Value:=Value/PfFak[k];
        Result:=true;
        end;
      end
    end
  else begin
    Value:=0; Result:=true;
    end;
  end;

(* ----------FUNCTION--FloatToEng---------------------------------------- *)
procedure FloatToEng (Value   : double; DecSep  : char; n : integer;
                      var Mant : string; var Exp : integer);
// Value   = Wert
// n       = Anzahl der signifikanten Stellen
// DecSep  = Dez.trenner
// Mant    = Mantisse mit n signif. Stellen
// Exp     = Exponent in 3er-Stufen
var
  f    : double;
  nd,v : integer;
begin
  Normal (Value,f,Exp);
  if n<=1 then n:=2;
  if n>8 then n:=8;
  nd:=0;
  if abs(Value)>=0.99999 then inc(nd);
  if abs(Value)>=9.9999 then inc(nd);
  if abs(Value)>=99.999 then inc(nd);
  v:=round(abs(Value)*FFak[n-nd]);
  if n<nd then v:=v*NFak[nd-n];
  if v=0 then begin
    Mant:='0'+DecSep;
    for nd:=1 to n do Mant:=Mant+'0';
    end
  else begin
    Mant:=IntToStr(v);
    if nd>0 then begin
      if n>nd then insert(DecSep,Mant,nd+1);
      end
    else Insert('0'+DecSep,Mant,1);
    if Value<0 then Insert('-',Mant,1);
    end;
  end;

(* Zahlen auf "Eng-Format" umsetzen (3er Potenzen) *)
function FloatToStrE (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;
(* n    = Anzahl der signifikanten Stellen *)
var
  s : string;
  k : integer;
begin
  if n=1 then Result:=IntToStr(round(Value))
  else begin
    if n=0 then begin
      FloatToEng (Value,DecSep,4,s,k);
      if (Pos(DecSep,s)>0) then begin  // automatisch
        while s[length(s)]='0' do delete(s,length(s),1);
        if s[length(s)]=DecSep then delete(s,length(s),1);
        end;
      end
    else FloatToEng (Value,DecSep,n,s,k);
    if k<>6 then Result:=s+'E'+IntToStr((k-6)*3)
    else Result:=s;
    end;
  end;

(* ----------FUNCTION--FloatToPlotString---------------------------------------- *)
(* Zahlen auf "Eng-Format" mit Vorstzen (m,k,..)  umsetzen *)
function FloatToPrefixStr (Value     : double;
                           DecSep    : char;
                           n         : integer;
                           Separator : string) : string;
(* n    = Anzahl der signifikanten Stellen *)
var
  s   : string;
  k   : integer;
begin
  if n=1 then Result:=IntToStr(round(Value))
  else begin
    FloatToEng (Value,DecSep,n,s,k);
    if (k<1) or (k>PfAnz) then Result:=s+'E'+IntToStr((k-6)*3)
    else if k<>6 then Result:=s+Separator+pfx[k] else Result:=s;
    end;
  end;

(* ----------FUNCTION--FloatToStringS---------------------------------------- *)
(* Zahlen auf "Sci-Format" umsetzen  *)
function FloatToStrS (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;
var
  k : integer;
  fs : TFormatSettings;
begin
  if n=1 then Result:=IntToStr(round(Value))
  else begin
    fs:=FormatSettings; fs.DecimalSeparator:=DecSep;
    if n=0 then begin
      Result:=FloatToStrF(Value,ffExponent,4,1,fs);
      k:=Pos('E',Result)-1;
      while Result[k]='0' do begin
        delete(Result,k,1); dec(k);
        end;
      if Result[k]=DecSep then delete(Result,k,1);
      end
    else Result:=FloatToStrF(Value,ffExponent,n,1,fs);
    k:=Pos('E',Result);
    if (k>0) then begin
      if (Result[k+1]='+') then Delete(Result,k+1,1);  // '+' im Exp. entfernen
      if (Result[k+1]='0') then Delete(Result,k,2);  // 'E0' entfernen
      end;
    end;
  end;


(* ----------FUNCTION--FloatToFixString---------------------------------------- *)
(* Zahlen auf "ffXixed-Format" umsetzen *)
function FloatToFixStr (Value     : double;
                        DecSep  : char;
                        n         : integer) : string;
(* n    = Anzahl der signifikanten Stellen + 100 * Stellen hinter dem Komma *)
var
  k   : integer;
  fs  : TFormatSettings;
  f   : double;
begin
  if n=1 then Result:=IntToStr(round(Value))
  else begin
    fs:=FormatSettings; fs.DecimalSeparator:=DecSep;
    if n=0 then Result:=FloatToStrF (Value,ffGeneral,1,1,fs)
    else begin
      n:=n mod 100; k:=n div 100;
      if k=0 then begin
        if abs(Value)<MinDouble then k:=1 else k:=round(aint(lg(abs(Value))))+1;
        if k<=0 then begin
          if k+n>0 then begin
            k:=n; inc(n);
            end;
          end
        else if n>k then k:=n-k
        else k:=0;
        end;
      if k>=0 then Result:=FloatToStrF (Value,ffFixed,n,k,fs)
      else Result:=FloatToStrF (Value,ffExponent,n,1,fs);
      end;
    end;
  end;

{------------------------------------------------------------------}
(* Winkel-Wert konvertieren *)
function StrToDeg (s     : string;
                   var v : double) : boolean;
var
 t      : String;
 x,y    : double;
 j,k    : integer;
begin
  j:=0;
  k:=pos(',',s); if k>0 then s[k]:='.';
  if length(s)<>0 then begin
    k:=pos(':',s);
    if k=0 then k:=pos('/',s);
    if k=0 then k:=pos('"',s);
    if k=0 then val(s,v,j)
    else begin (* Grad:Min *)
      val (copy(s,1,pred(k)),x,j); (* Grad *)
      if j=0 then begin
        t:=copy (s,succ(k),255);
        k:=pos(':',t);
        if k=0 then k:=pos('/',t);
        if k=0 then k:=pos(#$27,s);
        if k=0 then begin (* dez.Min *)
          if length(t)>0 then begin
            val (t,v,j);
            if (j=0) and (v<60.0) then begin
              if x>=0.0 then v:=x+v/60.0
              else v:=x-v/60.0;
              end
            else j:=1;
            end
          else v:=x;
          end
        else begin (* Sek. *)
          val (copy (t,1,pred(k)),y,j);  (* Min *)
          if (j=0) and (y<60.0) then begin
            t:=copy (t,succ(k),255);
            if length(t)>0 then begin
              val (t,v,j); (* Sek. *)
              if (j=0) and (v<60.0) then begin
                if x>=0.0 then v:=x+(y+v/60.0)/60.0
                else v:=x-(y-v/60.0)/60.0;
                end
              else j:=1;
              end
            else begin
              if x>=0.0 then v:=x+y/60.0
              else v:=x-y/60.0;
              end;
            end
          else j:=1;
          end;
        end;
      end;
    end
  else v:=0.0;
  StrToDeg:=j=0;
  end;

(* Winkel in String mit vorgebenem Format umsetzen
   Stellen: w : gesamt
            d : hinter Komma (einschl. Minuten und Sekunden
   Typ : m = fmDezGrad  : Grad,xxx
             fmDezMin   : Grad:Minuten,xx
             fmSekunden : Grad:Minuten:Sekunden,xxx *)
function DegToStr(v      : double;
                  Format : TAngleFormat;
                  DecSep : char;
                  w,d    : integer) : string;
var
  s,t   : String;
  x,y,z : extended;
  i,n   : integer;
  fs    : TFormatSettings;
begin
  x:=int(abs(v)); y:=60.0*frac(abs(v));
  fs:=FormatSettings; fs.DecimalSeparator:=DecSep;
  case Format of
  fmDezGrad : Result:=FloatToStrF(v,ffFixed,w,d,fs); //str(v:w:d,s);
  fmDezMin : begin (* Grad:Min *)
      t:=FloatToStrF(y,ffFixed,w-2,d-2);  // Minuten
      if copy(t,1,2)='60' then begin
        y:=0.0; x:=x+1.0;
        t:=FloatToStrF(y,ffFixed,w-2,d-2);  // Minuten
        end;
      if y<10 then t:='0'+t;
      if v<0.0 then x:=-x;
      Result:=FloatToStrF(x,ffFixed,2,0)+':'+t; // Grade
      end;
  fmSekunden : begin (* Grad:Min:Sek *)
      z:=int(y); y:=60.0*frac(y);
      t:=FloatToStrF(y,ffFixed,w-4,d-4);  // Sekunden
      if copy(t,1,2)='60' then begin
        y:=0.0; z:=z+1.0;
        if z>=60.0 then begin
          z:=0.0; x:=x+1.0;
          end;
        t:=FloatToStrF(y,ffFixed,w-4,d-4);  // Sekunden
        end;
      if y<10 then t:='0'+t;
      s:=FloatToStrF(z,ffFixed,2,0);  // Minuten
      if z<10 then s:='0'+s;
      t:=s+':'+t;
      if v<0.0 then x:=-x;
      Result:=FloatToStrF(x,ffFixed,2,0)+':'+t; // Grade
      end;
    end;
  end;

(* ----------Initialisierung-------------------------------------------- *)
begin
  (* Initialisierung der Konstanten *)
  ln10:=ln(10); ln2:=ln(2); sqrt2:=sqrt(2);
  end.
